/**
  *  \file planetscentral/host/cmdsched.cc
  *  \brief Host: SCHEDULE commands
  *
  *  A game can have multiple schedules. Each schedule can have an
  *  expiration condition (time or turn) after which it is dropped.
  *  This is used to model rules
  *  - "thrice a week until turn 25, then twice a week"
  *  - "pause until <date>"
  */

#include "planetscentral/host/cmdsched.h"
#include "planetscentral/host/schedule.h"
#include "planetscentral/host/db.h"
#include "planetscentral/host/lock.h"
#include "int/args.h"
#include "planetscentral/host/host.h"
#include "planetscentral/osutil.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "cpluslib/format.h"
#include "planetscentral/host/cron.h"
#include "int/array.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/game.h"

/** Parse commands for a SCHEDULESET/ADD/MOD command.
    \param args [in] Arguments received from user
    \param sched [in/out] Schedule to fill in
    \param hadDaytime [out] Reports whether the schedule contained a DAYTIME option
    \return true on success, false on failure */
bool
parseSchedule(IntArgBlock& args, Schedule& sched, bool& hadDaytime)
{
    while (args.getNumArgs() > 0) {
        string_t keyword = strUCase(toStringResult(args.getNext()));
        if (keyword == "STOP") {
            sched.setType(Schedule::Stopped);
        } else if (keyword == "WEEKLY") {
            args.checkArgumentCountAtLeast(1);
            sched.setType(Schedule::Weekly);
            sched.setWeekDays(SmallSet<int8_t>::fromInteger(toIntResult(args.getNext())) & SmallSet<int8_t>::allUpTo(6));
        } else if (keyword == "DAILY") {
            args.checkArgumentCountAtLeast(1);
            sched.setType(Schedule::Daily);
            sched.setInterval(toIntResult(args.getNext()));
        } else if (keyword == "ASAP") {
            sched.setType(Schedule::Quick);
        } else if (keyword == "MANUAL") {
            sched.setType(Schedule::Manual);
        } else if (keyword == "DAYTIME") {
            args.checkArgumentCountAtLeast(1);
            sched.setDaytime(toIntResult(args.getNext()));
            hadDaytime = true;
        } else if (keyword == "EARLY") {
            sched.setHostEarly(true);
        } else if (keyword == "NOEARLY") {
            sched.setHostEarly(false);
        } else if (keyword == "DELAY") {
            args.checkArgumentCountAtLeast(1);
            sched.setHostDelay(toIntResult(args.getNext()));
        } else if (keyword == "LIMIT") {
            args.checkArgumentCountAtLeast(1);
            sched.setHostLimit(toIntResult(args.getNext()));
        } else if (keyword == "UNTILTURN") {
            args.checkArgumentCountAtLeast(1);
            sched.setCondition(Schedule::TurnCondition, toIntResult(args.getNext()));
        } else if (keyword == "UNTILTIME") {
            args.checkArgumentCountAtLeast(1);
            sched.setCondition(Schedule::TimeCondition, toIntResult(args.getNext()));
        } else if (keyword == "FOREVER") {
            sched.setCondition(Schedule::NoCondition, 0);
        } else {
            return false;
        }
    }
    return true;
}

/** Pack a schedule into an object.
    \param h "game:GID:schedule:ID" object */
IntValue*
packSchedule(DbHash h)
{
    /* @type HostSchedule
       Describes a schedule.
       This is essentially the content of {game:$GID:schedule:$SID}.

       @key type:Int        (0=stopped / 1=weekly / 2=daily / 3=asap / 4=manual)
       @key weekdays:Int    (if weekly: sum of 1<<0(Sunday) .. 1<<6(Saturday))
       @key interval:Int    (if daily: days between host)
       @key daytime:Int     (if weekly/daily: preferred day time (minutes))
       @key hostEarly:Int   (if weekly/daily: host early when all turns are in; defaults to true if asap)
       @key hostDelay:Int   (if hostEarly or asap: delay after last turn submission, minutes, default: 30)
       @key hostLimit:Int   (skip next host if host terminated more than this many minutes after start, default: 360 = 6 hours)
       @key condition:Int   (0=none / 1=turn / 2=time)
       @key condTurn:Int    (if turn: drop this condition at turn N)
       @key condTime:Time   (if time: drop this condition at time)
       @c        [- tempLimit (time limit for temporary turns)] */
    Ptr<IntArrayData> a = new IntArrayData();
    StringList_t list;
    h.getAll(list);
    for (uint32_t i = 0; i < list.size(); ++i) {
        // FIXME: this packs the condTime in raw format, which is an error
        // when Host.TimeScale is not 60 (should be passed through getUserTimeFromTime).
        a->content.pushNew(makeStringValue(list[i]));
    }
    a->addDimension(a->content.size());
    return new IntArray(a);
}

/** SCHEDULEADD or SCHEDULESET command.
    \param conn Connection
    \param args Arguments received from user
    \param add true for SCHEDULEADD, false for SCHEDULESET */
static void
doSchedule(HostConnection& conn, IntArgBlock& args, bool add)
{
    // Parse arguments
    args.checkArgumentCountAtLeast(2);
    int32_t gameId = toIntResult(args.getNext());
    bool hadDaytime = false;
    Schedule sched;
    if (!parseSchedule(args, sched, hadDaytime)) {
        conn.sendError("400 Syntax error");
        return;
    }

    // Lock game
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Database work
    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, GAME_ROOT);
    Game game(root, gameId);
    conn.checkPermission(game, Game::ConfigPermission);

    // If the daytime was not set, try to derive it from the existing schedule.
    DbSubtree sroot(game.getSchedule());
    bool anySchedule = sroot.stringListKey("list").size() > 0;
    if (!hadDaytime) {
        if (anySchedule) {
            string_t currentSchedule = sroot.stringListKey("list")[0];
            sched.setDaytime(sroot.hashKey(currentSchedule).getIntField("daytime"));
        } else {
            sched.setDaytime(pickDayTime(root));
        }
    }

    // Process command.
    if (add || !anySchedule) {
        string_t newSchedule = format("%d", ++sroot.intKey("lastId"));
        sched.saveTo(sroot.hashKey(newSchedule));
        sroot.stringListKey("list").pushFront(newSchedule);
    } else {
        sched.saveTo(sroot.hashKey(sroot.stringListKey("list")[0]));
    }
    game.removeConfig("hostRunNow");
    game.setConfigInt("lastScheduleChange", getCurrentTime());
    game.scheduleChanged().set(1);
    conn.sendSuccess("OK");
    cron.handleGameChange(gameId);
}

/* @q SCHEDULEADD game:GID [scheduleParams...] (Host Command)
   Add new schedule for a game.
   This schedule will be the active schedule until it expires.

   The %scheduleParams are a list of keywords and parameters:
   - STOP (schedule type: do not run host)
   - WEEKLY n:{@type Int} (schedule type: host runs on fixed weekdays. %n is the bitfield of days; 0=Sunday, 1=Monday, ... 6=Saturday)
   - DAILY n:{@type Int} (schedule type: host runs every %n days)
   - ASAP (schedule type: host runs when all turns are in)
   - MANUAL (schedule type: host is run manually)
   - EARLY (acceleration: host runs early if all turns are in)
   - NOEARLY (acceleration: host does not run early)
   - DELAY n:{@type Int} (acceleration: minutes after last turn file to wait until host runs)
   - DAYTIME n:{@type Int} (for weekly/daily schedules: runs %n minutes after midnight)
   - LIMIT n:{@type Int} (for weekly/daily schedules: maximum minutes acceptable delay after which a host date is moved)
   - UNTILTURN n:{@type Int} (expiration: schedule expires after turn %n)
   - UNTILTIME n:{@type Time} (expiration: schedule expires at time %n)
   - FOREVER (expiration: schedule does not expire)

   Defaults are STOP, EARLY, DELAY 30 (half an hour), LIMIT 360 (six hours), FOREVER.
   The default DAYTIME is assigned individually for each game (see {game:hours}).

   Permissions: config-access to game.

   @uses game:$GID:schedule:list, game:$GID:schedule:$SID, game:$GID:schedule:lastId
   @see SCHEDULESET, SCHEDULEMOD, SCHEDULELIST */
void
doScheduleAdd(HostConnection& conn, IntArgBlock& args)
{
    doSchedule(conn, args, true);
}

/* @q SCHEDULESET game:GID [scheduleParams...] (Host Command)
   Replace schedule.
   See {SCHEDULEADD} for a description of the %scheduleParams.

   The difference between this command and {SCHEDULEMOD} is that
   this one drops the front schedule and builds a new one from scratch.
   If there is no schedule, this command behaves like {SCHEDULEADD}.

   Permissions: config-access to game.

   @uses game:$GID:schedule:list, game:$GID:schedule:$SID, game:$GID:schedule:lastId
   @see SCHEDULEADD, SCHEDULEMOD, SCHEDULELIST */
void
doScheduleSet(HostConnection& conn, IntArgBlock& args)
{
    doSchedule(conn, args, false);
}

/* @q SCHEDULEMOD game:GID [scheduleParams...] (Host Command)
   Modify schedule.
   See {SCHEDULEADD} for a description of the %scheduleParams.

   The difference between this command and {SCHEDULESET} is that
   this one edits the existing schedule in place and fails if there isn't one.

   Permissions: config-access to game.

   @err 410 Game does not have a schedule
   @uses game:$GID:schedule:list, game:$GID:schedule:$SID, game:$GID:schedule:lastId
   @see SCHEDULEADD, SCHEDULESET, SCHEDULELIST */
void
doScheduleModify(HostConnection& conn, IntArgBlock& args)
{
    // Parse arguments
    args.checkArgumentCountAtLeast(1);
    int32_t gameId = toIntResult(args.getNext());

    // Lock game
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Database work
    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, GAME_ROOT);
    Game game(root, gameId);
    conn.checkPermission(game, Game::ConfigPermission);

    // Load existing schedule
    DbSubtree sroot(game.getSchedule());
    if (sroot.stringListKey("list").size() == 0) {
        conn.sendError("410 Game does not have a schedule");
        return;
    }
    string_t scheduleName = sroot.stringListKey("list")[0];

    Schedule sched;
    bool hadDaytime = false;
    sched.loadFrom(sroot.hashKey(scheduleName));
    if (!parseSchedule(args, sched, hadDaytime)) {
        conn.sendError("400 Syntax error");
        return;
    }

    // Save it back
    game.removeConfig("hostRunNow");
    sched.saveTo(sroot.hashKey(scheduleName));
    game.setConfigInt("lastScheduleChange", getCurrentTime());
    game.scheduleChanged().set(1);
    cron.handleGameChange(gameId);
    conn.sendSuccess("OK");
}

/* @q SCHEDULELIST game:GID (Host Command)
   List schedules of game.

   Permissions: read-access to game.

   @retval HostSchedule[] List of active schedules */
void
doScheduleList(HostConnection& conn, IntArgBlock& args)
{
    // Parse arguments
    args.checkArgumentCount(1);
    int32_t gameId = toIntResult(args.getNext());

    // Database work
    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, GAME_ROOT);
    Game game(root, gameId);
    conn.checkPermission(game, Game::ReadPermission);

    // Create result
    DbSubtree sroot(game.getSchedule());
    StringList_t schedules;
    sroot.stringListKey("list").getAll(schedules);
    Ptr<IntArrayData> a = new IntArrayData();
    for (uint32_t i = 0; i < schedules.size(); ++i) {
        a->content.pushNew(packSchedule(sroot.hashKey(schedules[i])));
    }
    a->addDimension(a->content.size());
    IntArray result(a);
    conn.sendObject(&result);
}

/* @q SCHEDULEDROP game:GID (Host Command)
   Drop first schedule.
   If the game has multiple schedules, the following schedule will kick in.
   Otherwise, the game will stop hosting.

   Permissions: config-access to game.

   @uses game:$GID:schedule:list */
void
doScheduleDrop(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    int32_t gameId = toIntResult(args.getNext());

    // Database work
    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, GAME_ROOT);
    Game game(root, gameId);
    conn.checkPermission(game, Game::ConfigPermission);

    DbSubtree sroot(game.getSchedule());
    string_t removedSchedule = sroot.stringListKey("list").popFront();
    if (!removedSchedule.empty()) {
        sroot.hashKey(removedSchedule).remove();
        game.removeConfig("hostRunNow");
    }
    game.scheduleChanged().set(1);
    game.setConfigInt("lastScheduleChange", getCurrentTime());
    cron.handleGameChange(gameId);
    conn.sendSuccess("OK");
}

/* @q SCHEDULESHOW game:GID [TIMELIMIT n:Int] [TURNLIMIT n:Int] (Host Command)
   Lists all future host dates.
   Computes future hosting times until either the time limit or the turn limit has been reached,
   or no more host times can be computed.

   Permissions: read-access to game.

   @retval IntList list of {@type Time}s of future host runs
   @rettype Time */
void
doScheduleShow(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    int32_t gameId = toIntResult(args.getNext());
    int32_t turnLimit = 0;
    int32_t timeLimit = 0;
    while (args.getNumArgs() > 0) {
        string_t keyword = strUCase(toStringResult(args.getNext()));
        if (keyword == "TIMELIMIT") {
            args.checkArgumentCountAtLeast(1);
            timeLimit = getCurrentTime() + toIntResult(args.getNext());
        } else if (keyword == "TURNLIMIT") {
            args.checkArgumentCountAtLeast(1);
            turnLimit = toIntResult(args.getNext());
        } else {
            conn.sendError("400 Syntax error");
            return;
        }
    }

    // Database work
    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, GAME_ROOT);
    Game game(root, gameId);
    conn.checkPermission(game, Game::ReadPermission);

    Ptr<IntArrayData> a = new IntArrayData();

    // Figure out last host date
    int32_t lastHost = game.getConfigInt("lastHostTime");
    int32_t turn     = game.getConfigInt("turn");
    if (lastHost == 0) {
        // Host never run, so pretend we're hosting now.
        lastHost = getCurrentTime();
        ++turn;
        a->content.pushNew(makeIntValue(getUserTimeFromTime(lastHost)));
    }

    // Process all schedules
    DbSubtree sroot(game.getSchedule());
    DbStringList schedules(sroot.stringListKey("list"));
    int32_t numSchedules = schedules.size();
    int32_t currentSchedule = 0;
    while (currentSchedule < numSchedules
           && turnLimit > int32_t(a->content.size())
           && (timeLimit == 0 || timeLimit > lastHost)
           && lastHost != 0)
    {
        // Process one schedule
        // FIXME: this does not seem to handle properly transitions from
        // one DAILY/WEEKLY schedule to another with a time condition.
        Schedule sched;
        sched.loadFrom(sroot.hashKey(schedules[currentSchedule]));
        while (turnLimit > int32_t(a->content.size())
               && (timeLimit == 0 || timeLimit > lastHost)
               && lastHost != 0
               && !sched.isExpired(turn, lastHost))
        {
            lastHost = sched.getNextHost(lastHost);
            ++turn;
            a->content.pushNew(makeIntValue(getUserTimeFromTime(lastHost)));
        }
        ++currentSchedule;
    }
    a->addDimension(a->content.size());

    // Return result
    IntArray result(a);
    conn.sendObject(&result);
}
