/**
  *  \file planetscentral/talk/cmdforum.cc
  *  \brief Talk: FORUM commands
  */

#include <stdexcept>
#include "planetscentral/talk/cmdforum.h"
#include "int/args.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/forum.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/talkconnection.h"
#include "planetscentral/talk/db.h"
#include "int/array.h"

const ListParams::SortParams ListParams::threadSortKeys[] =
{
    { "FIRSTPOST", THREAD_ROOT "*:header->firstpost", false },
    { "FORUM",     THREAD_ROOT "*:header->forum",     false },
    { "LASTPOST",  THREAD_ROOT "*:header->lastpost",  false },
    { "LASTTIME",  THREAD_ROOT "*:header->lasttime",  false },
    { "SUBJECT",   THREAD_ROOT "*:header->subject",   true },
    { 0, 0, false },
};

const ListParams::SortParams ListParams::postSortKeys[] =
{
    { "AUTHOR",    MESSAGE_ROOT "*:header->author",   true },
    { "EDITTIME",  MESSAGE_ROOT "*:header->edittime", false },
    { "SUBJECT",   MESSAGE_ROOT "*:header->subject",  true },
    { "THREAD",    MESSAGE_ROOT "*:header->thread",   false },
    { "TIME",      MESSAGE_ROOT "*:header->time",     false },
    { 0, 0, false },
};

const ListParams::SortParams ListParams::forumSortKeys[] =
{
    { "KEY",       FORUM_ROOT "*:header->key",        true },
    { "LASTPOST",  FORUM_ROOT "*:header->lastpost",   false },
    { "LASTTIME",  FORUM_ROOT "*:header->lasttime",   false },
    { "NAME",      FORUM_ROOT "*:header->name",       true },
    { 0, 0, false },
};

const ListParams::SortParams ListParams::pmSortKeys[] =
{
    { "AUTHOR",    PM_ROOT "*:header->author",  true },
    { "SUBJECT",   PM_ROOT "*:header->subject", true },
    { "TIME",      PM_ROOT "*:header->time",    false },
    // { "FLAGS",  PM_ROOT "*:header->flags/$USER" },
    { 0, 0, false },
};


ListParams::ListParams()
    : id(0), start(0), count(0), item(0), mode(WantAll), sortKey(0)
{ }

void
ListParams::parse(IntArgBlock& args, bool withId, const SortParams* sortParams)
{
    /* @page pcc:talk:listparams, List Parameters
       @in pcc:resp
       Various {@group Talk Command}s return lists of things.
       These commands accept a common set of options to limit or modify the result.

       - <tt>LIMIT start count</tt>: return at most %count elements, starting at the %start'th (returns a list)
       - <tt>SIZE</tt>: return number of elements (integer)
       - <tt>CONTAINS n</tt>: returns 1 if the list contains element %n, otherwise 0 (integer)
       - <tt>SORT key</tt>: if a list is returned, sort it by the given key.
         Valid sort keys depend on the command.
         The result is always sorted in ascending order. */
    if (withId) {
        args.checkArgumentCountAtLeast(1);
        id = toIntResult(args.getNext());
    }
    while (args.getNumArgs() > 0) {
        string_t key = strUCase(toStringResult(args.getNext()));
        if (key == "LIMIT") {
            args.checkArgumentCountAtLeast(2);
            start = toIntResult(args.getNext());
            count = toIntResult(args.getNext());
            mode = WantRange;
        } else if (key == "SIZE") {
            mode = WantSize;
        } else if (key == "CONTAINS") {
            args.checkArgumentCountAtLeast(1);
            item = toIntResult(args.getNext());
            mode = WantMemberCheck;
        } else if (key == "SORT") {
            args.checkArgumentCountAtLeast(1);
            string_t key = strUCase(toStringResult(args.getNext()));
            sortKey = 0;
            if (sortParams != 0) {
                for (int i = 0; sortParams[i].name != 0; ++i) {
                    if (sortParams[i].name == key) {
                        sortKey = &sortParams[i];
                        break;
                    }
                }
            }
            if (!sortKey) {
                throw std::runtime_error("400 Invalid sort key");
            }
        } else {
            throw std::runtime_error("400 Invalid option");
        }
    }
}

void
ListParams::exec(TalkConnection& conn, DbIntSet set)
{
    if (mode == WantSize) {
        IntIntIntValue result(set.size());
        conn.sendObject(&result);
    } else if (mode == WantMemberCheck) {
        IntIntIntValue result(set.contains(item));
        conn.sendObject(&result);
    } else {
        DbSort sort(set.sort());
        if (mode == WantRange) {
            sort.limit(start, count);
        }
        if (sortKey != 0) {
            sort.by(sortKey->sortKey);
            if (sortKey->lexicographically) {
                sort.sortLexicographical();
            }
        }
        std::auto_ptr<IntValue> result(sort.getResult());
        conn.sendObject(result.get());
    }
}

int32_t
ListParams::getId()
{
    return id;
}

static void
configureForum(Forum& f, IntArgBlock& args)
{
    while (args.getNumArgs() > 0) {
        args.checkArgumentCountAtLeast(2);
        string_t key(toStringResult(args.getNext()));
        string_t value(toStringResult(args.getNext()));
        if (key == "parent") {
            f.setParent(value);
        } else if (key == "newsgroup") {
            f.setNewsgroup(value);
        } else {
            f.header().setField(key, value);
        }
    }
}

/* @q FORUMADD [key:Str value:Str ...] (Talk Command)
   Create forum.
   The command is followed by %key/%value pairs that configure the forum, see {FORUMSET}.

   Permissions: admin.

   @retval FID new forum Id
   @uses forum:$FID:header, forum:all */
void
doForumAdd(TalkConnection& conn, IntArgBlock& args)
{
    // Allocate FID
    ReconnectingGuard g(database_rc);
    conn.checkAdmin();
    int32_t newFid = Forum::allocateForum(database_connection);

    // Create forum
    Forum f(database_connection, newFid);
    DbSubtree(database_connection, FORUM_ROOT).intSetKey("all").add(newFid);
    DbHashIntField ctime(f.creationTime());
    if (!ctime.exists()) {
        ctime.set(getTime());
    }

    // Configure it
    configureForum(f, args);

    // Result
    IntIntIntValue iv(newFid);
    conn.sendObject(&iv);
}

/* @q FORUMSET forum:FID [key:Str value:Str ...] (Talk Command)
   Configure forum.
   Valid keys are the same as keys in {forum:$FID:header}; sensible values listed here:
   - %name ({@type Str|string}, forum name)
   - %parent ({@type GRID}, containing group)
   - %description ({@type TalkText}, description/subtitle)
   - %newsgroup ({@type Str|string}, NNTP newsgroup name)
   - %readperm, %writeperm, %answerperm, %deleteperm ({@type TalkPerm}, configure permissions)
   - %key ({@type Str|string}, sort key)

   Permissions: none.

   @uses forum:$FID:header, group:$GRID:forums, forum:newsgroups
   @err 404 Not found
   @argtype Str
   @argtype GRID
   @argtype TalkText
   @argtype TalkPerm */
void
doForumSet(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    int32_t fid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    conn.checkAdmin();
    Forum f(database_connection, fid);
    if (!f.exists()) {
        conn.sendError("404 Not found");
    } else {
        configureForum(f, args);
        conn.sendSuccess("OK");
    }
}

/* @q FORUMGET forum:FID key:Str (Talk Command)
   Get forum property.
   Valid keys are the same as keys in {forum:$FID:header}.

   Permissions: none.

   The returned value is taken directly from the forum header.

   @uses forum:$FID:header
   @err 404 Not found
   @rettype Str
   @rettype GRID
   @rettype TalkText
   @rettype TalkPerm
   @rettype Time
   @rettype Int */
void
doForumGet(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    int32_t fid = toIntResult(args.getNext());
    string_t key = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Forum f(database_connection, fid);
    if (!f.exists()) {
        conn.sendError("404 Not found");
    } else {
        std::auto_ptr<IntValue> result(f.header().getFieldRaw(key));
        conn.sendObject(result.get());
    }
}

/* @q FORUMSTAT forum:FID (Talk Command)
   Get forum information.

   Permissions: none.

   Note that this command uses the rendering configuration set with {RENDEROPTION}
   to render the forum description.

   @retval TalkForumInfo information about forum
   @see FORUMMSTAT
   @uses forum:$FID:header */
void
doForumStat(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    int32_t fid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Forum f(database_connection, fid);
    if (!f.exists()) {
        conn.sendError("404 Not found");
    } else {
        std::auto_ptr<IntValue> result(f.describe(conn.getRenderState()));
        conn.sendObject(result.get());
    }
}

/* @q FORUMMSTAT forum:FID... (Talk Command)
   Get multiple forums information.

   Permissions: none.

   Note that this command uses the rendering configuration set with {RENDEROPTION}
   to render the forum descriptions.

   @retval TalkForumInfo[] information about forum
   @see FORUMSTAT
   @uses forum:$FID:header */
void
doForumMStat(TalkConnection& conn, IntArgBlock& args)
{
    // Do it
    ReconnectingGuard g(database_rc);
    Ptr<IntArrayData> ad(new IntArrayData());
    while (args.getNumArgs() > 0) {
        int32_t fid = toIntResult(args.getNext());
        Forum f(database_connection, fid);
        if (!f.exists()) {
            conn.sendError("404 Not found");
            return;
        }
        ad->content.pushNew(f.describe(conn.getRenderState()));
    }
    ad->addDimension(ad->content.size());

    // Send result
    IntArray arr(ad);
    conn.sendObject(&arr);
}

/* @q FORUMPERMS forum:FID [perm:Str ...] (Talk Command)
   Get forum permissions.
   For each given permission name, checks whether the user has the respective privilege.
   - %read (read postings)
   - %write (create new threads)
   - %answer (answer to a posting)
   - %delete (delete postings)
   The returned value is an integer with each bit corresponding to a privilege.
   For example, "FORUMPERMS 1 answer write" returns the "answer" permission in bit 0, the "write" permission in bit 1.

   Permissions: none (everyone can execute this command).

   @err 404 Not found
   @retval Int permissions
   @uses forum:$FID:header */
void
doForumPerms(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    int32_t fid = toIntResult(args.getNext());
    int32_t result = 0;
    int32_t mask = 1;

    // Do it
    ReconnectingGuard g(database_rc);
    Forum f(database_connection, fid);
    if (!f.exists()) {
        conn.sendError("404 Not found");
    } else {
        while (args.getNumArgs() > 0) {
            string_t key = toStringResult(args.getNext()) + "perm";
            string_t str = f.header().getStringField(key);
            if (conn.hasPrivilege(str)) {
                result |= mask;
            }
            mask <<= 1;
        }
        IntIntIntValue iv(result);
        conn.sendObject(&iv);
    }
}

/* @q FORUMSIZE forum:FID (Talk Command)
   Get forum size.

   Permissions: none.

   @retkey threads:Int (Number of threads)
   @retkey stickythreads:Int (Number of sticky threads)
   @retkey messages:Int (Number of messages)

   @err 404 Not found
   @uses forum:$FID:messages, forum:$FID:threads, forum:$FID:stickythreads */
void
doForumSize(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    int32_t fid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Forum f(database_connection, fid);
    if (!f.exists()) {
        conn.sendError("404 Not found");
    } else {
        conn.checkPrivilege(f.readPermissions().get());
        Ptr<IntArrayData> ad(new IntArrayData());
        ad->content.pushNew(makeStringValue("threads"));
        ad->content.pushNew(makeIntValue(f.topics().size()));
        ad->content.pushNew(makeStringValue("stickythreads"));
        ad->content.pushNew(makeIntValue(f.stickyTopics().size()));
        ad->content.pushNew(makeStringValue("messages"));
        ad->content.pushNew(makeIntValue(f.messages().size()));
        ad->addDimension(ad->content.size());
        IntArray a(ad);
        conn.sendObject(&a);
    }
}

/* @q FORUMLSTHREAD forum:FID [listParameters...] (Talk Command)
   Query list of threads in a forum.

   The list can be accessed in different ways, see {pcc:talk:listparams|listParameters}.
   Valid sort keys for threads are:
   - firstpost
   - forum
   - lastpost
   - lasttime
   - subject

   Permissions: none (everyone can execute this command).
   @c FIXME: is this a good idea?

   If Ids are returned, these are {@type TID|thread Ids}.

   @err 404 Not found
   @rettype Any
   @rettype TID
   @uses forum:$FID:threads */
void
doForumListThreads(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    ListParams parms;
    parms.parse(args, true, ListParams::threadSortKeys);

    // Do it
    ReconnectingGuard g(database_rc);
    Forum f(database_connection, parms.getId());
    if (!f.exists()) {
        conn.sendError("404 Not found");
    } else {
        parms.exec(conn, f.topics());
    }
}

/* @q FORUMLSSTICKY forum:FID [listParameters...] (Talk Command)
   Query list of sticky threads in a forum.

   The list can be accessed in different ways, see {pcc:talk:listparams|listParameters}.
   Valid sort keys for threads are:
   - firstpost
   - forum
   - lastpost
   - lasttime
   - subject

   Permissions: none (everyone can execute this command).
   @c FIXME: is this a good idea?

   If Ids are returned, these are {@type TID|thread Ids}.

   @err 404 Not found
   @rettype Any
   @rettype TID
   @uses forum:$FID:stickythreads */
void
doForumListStickyThreads(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    ListParams parms;
    parms.parse(args, true, ListParams::threadSortKeys);

    // Do it
    ReconnectingGuard g(database_rc);
    Forum f(database_connection, parms.getId());
    if (!f.exists()) {
        conn.sendError("404 Not found");
    } else {
        parms.exec(conn, f.stickyTopics());
    }
}

/* @q FORUMLSPOST forum:FID [listParameters...] (Talk Command)
   Query list of postings in a forum.

   The list can be accessed in different ways, see {pcc:talk:listparams|listParameters}.
   Valid sort keys for postings are:
   - author
   - edittime
   - subject
   - thread
   - time

   Permissions: none (everyone can execute this command).
   @c FIXME: is this a good idea?

   If Ids are returned, these are {@type MID|message Ids}.

   @err 404 Not found
   @rettype Any
   @rettype MID
   @uses forum:$FID:messages */
void
doForumListPosts(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    ListParams parms;
    parms.parse(args, true, ListParams::forumSortKeys);

    // Do it
    ReconnectingGuard g(database_rc);
    Forum f(database_connection, parms.getId());
    if (!f.exists()) {
        conn.sendError("404 Not found");
    } else {
        parms.exec(conn, f.messages());
    }
}
